% Robust control of an exothermic reaction.

% Load parameters for problem.
nompars = cstrparam();
nompars.tfinal = 600;
nompars.colloc = 10;
nompars.plantSteps = 20;

% Disturbance.
nompars.A = 0.002;
nompars.w = 0.4;

% Different sample times.
Deltas = [8, 12];
pars = struct();
for i = 1:length(Deltas)
    f = sprintf('d%d', Deltas(i));
    pars.(f) = nompars;
    pars.(f).Delta = Deltas(i);
end
pars.d12.colloc = 24; % Need more points for longer sample time.
fields = fieldnames(pars);

% Prediction Horizon is same.
predHorizon = 360;
for i = 1:length(fields)
    f = fields{i};
    pars.(f).N = round(predHorizon/pars.(f).Delta);
end

% Run nominal (central path) MPC using sample time of 12. Then resample.
reftraj = nominalmpc(pars.d12);


% Run tube-based MPC multiple times.
results = struct();
for i = 1:length(fields)
    f = fields{i};
    pars.(f).reftraj = resample_nominal(reftraj, pars.(f).Delta);
    results.(f) = tubempc(pars.(f));
end

% Make plot.
figure();
Nrows = 3;
Ncols = length(fields);
ylabels = {'concentration', 'temperature', 'coolant flow'};
ylims = {[0, 1], [0.3, 0.8], [-0.1, 1.4]};
for j = 1:Ncols
    f = fields{j};
    for i = 1:Nrows
        subplot(Nrows, Ncols, (i - 1)*Ncols + j);
        hold('on');
        if i == 3
            stairs(results.(f).t, results.(f).u([1:end,end]), '-k');
            stairs(reftraj.t, reftraj.u([1:end,end]), '--b');
        else
            plot(results.(f).t, results.(f).x(i,:), '-k', ...
                 reftraj.t, reftraj.x(i,:), '--b');
        end
        xlim([0, 500]);
        ylim(ylims{i});
        if i == 1
            title(sprintf('Delta = %s', f(2:end)));
        elseif i == Nrows
            xlabel('Time');
        end
        if j == 1
            ylabel(ylabels{i});
        end
    end
end

