% Comparison of different prior update strategies for a nonlinear system.
% Also compares EKF.
mpc = import_mpctools();

Nx = 2;
Ny = 1;
Nt = 10; % MHE horizon.

k1 = 0.16;
k2 = 0.0064;

Delta = 0.1;

% Define system model.
ode = @(x) [-2*k1*x(1)^2 + 2*k2*x(2); ...
            k1*x(1)^2 - k2*x(2)];
measurement = @(x) x(1) + x(2);

sim = mpc.getCasadiIntegrator(ode, Delta, [Nx], {'x'}, 'funcname', 'sim');
f = mpc.getCasadiFunc(ode, [Nx], {'x'}, 'funcname', 'f', ...
                      'rk4', true(), 'Delta', Delta);
h = mpc.getCasadiFunc(measurement, [Nx], {'x'}, 'funcname', 'h');

% Simulate the system.
Nsim = 200;
xsim = NaN(Nx, Nsim + 1);
xsim(:,1) = [3; 1];
ysim = NaN(Ny, Nsim + 1);
for i = 1:(Nsim + 1)
    ysim(:,i) = measurement(xsim(:,i));
    if i <= Nsim
        xsim(:,i + 1) = full(sim(xsim(:,i)));
    end
end

% Choose MHE objective function.
Q = diag([1e-4, 1e-2]);
R = diag([1e-2]);
P0 = diag([1, 1]);

Qinv = mpc.spdinv(Q);
Rinv = mpc.spdinv(R);
P0inv = mpc.spdinv(P0);

l = mpc.getCasadiFunc(@(w, v) w'*Qinv*w + v'*Rinv*v, [Nx, Ny], {'w', 'v'}, ...
                      'funcname', 'l');

par = struct('x0bar', [0.1; 4.5], 'Pinv', P0inv); % Initial prior.

% Choose other parameters.
lb = struct('x', zeros(Nx, 1));
N = struct('x', Nx, 'y', Ny, 't', Nt);
kwargs = struct('f', f, 'h', h, 'l', l, 'N', N, 'lb', lb, 'par', par, ...
                'y', ysim(:,1:(N.t + 1)), 'verbosity', 0);

% Simulate different prior updates.
updates = {'filtering', 'smoothing'};
mhes = struct();
xs = struct('actual', xsim);
ys = struct('actual', ysim);
for k = 1:length(updates)
    key = updates{k};
    fprintf('Simulating MHE with %s update.\n', key);
    mhe = mpc.nmhe('priorupdate', key, '**', kwargs);
    mhe.fix_truncated_x = true();

    xmhe = NaN(Nx, Nsim + 1);
    ymhe = NaN(Ny, Nsim + 1);
    for i = 0:Nsim
        if i <= Nt
            mhe.truncatehorizon(i);
        else
            mhe.newmeasurement(ysim(:,i + 1));
        end

        mhe.solve();
        if ~isequal(mhe.status, 'Solve_Succeeded')
            warning('Solver failed (status %s)!', mhe.status);
            break
        end
        mhe.saveestimate();
        xmhe(:,i + 1) = mhe.history(1).xhat(:,end);
        ymhe(:,i + 1) = mhe.history(1).yhat(:,end);
    end

    mhes.(key) = mhe;
    xs.(key) = xmhe;
    ys.(key) = ymhe;
end

% Use EKF. Assumes G = I.
fprintf('Simulating EKF\n');
xekf = NaN(Nx, Nsim + 1);
xekf(:,1) = par.x0bar;
P = P0;
yekf = NaN(Ny, Nsim + 1);
Afunc = f.factory('A', {'x'}, {'jac:f:x'});
Cfunc = h.factory('C', {'x'}, {'jac:h:x'});
for i = 1:(Nsim + 1)
    % Measurement.
    e = ysim(:,i) - full(h(xekf(:,i)));

    % Correction
    C = full(Cfunc(xekf(:,i)));
    L = (P*C')/(C*P*C' + R); % EKF gain
    P = P - L*C*P;

    xekf(:,i) = xekf(:,i) + L*e;
    yekf(:,i) = full(h(xekf(:,i)));

    % Advance forecast.
    if i <= Nsim
        A = full(Afunc(xekf(:,i)));
        xekf(:,i + 1) = full(f(xekf(:,i)));
        P = A*P*A' + Q;
    end
end
xs.ekf = xekf;
ys.ekf = yekf;

% Make a plot.
fprintf('Plotting.\n');
keys = {'actual', 'ekf', 'filtering', 'smoothing'};
colors = {'k', 'r', 'b', 'g'};
plotstyle = struct('t', (0:Nsim)*Delta, 'fig', figure(), 'marker', '', ...
                   'xnames', {{'p_a', 'p_b', 'p_{tot}'}}, ...
                   'unames', {{'log_{10}|e_a|', 'log_{10}|e_b|'}});
for i = 1:length(keys)
    k = keys{i};
    x = [xs.(k); ys.(k)];
    e = log10(abs(xs.(k) - xs.actual));
    mpc.mpcplot(x, e, 'legend', k, 'color', colors{i}, ...
                '**', plotstyle);
end

