function data = doukf(Nsim, clipping)
    % data = doukf(Nsim, clipping)
    %
    % Simulate the UKF with or without clipping.
    pars = batchreactor(Nsim);
    Nx = pars.Nx;
    Ny = pars.Ny;
    Nw = pars.Nw;
    Nv = pars.Nv;
    pars.Nz = Nx + Nw + Nv; % Augmented state is z = [x; w; v].
    Nz = pars.Nz;

    % Choose simulation parameters.
    Nsim = pars.Nsim;
    t = pars.t;
    x = pars.xsim;
    y = pars.ysim;
    Px = pars.P;
    Qw = pars.Q;
    Rv = pars.R;

    % Choose UKF weights.
    Nsigma = 1 + 2*Nz;
    pars.Nsigma = Nsigma;
    pars.lambda = 1; % From Vachhani et. al. (2006). Guarantees Pz > 0.
    pars.weights = [2*pars.lambda; ones(2*Nz, 1)]/(2*(Nz + pars.lambda));
    W = diag(pars.weights);

    % Preallocate variables.
    xhat = NaN(Nx, Nsim + 1);
    yhat = NaN(Ny, Nsim + 1);
    e = NaN(Ny, Nsim + 1);

    xhat(:,1) = pars.xhat0; % Poor initial guess
    zhat(1:Nx,1) = pars.xhat0;

    Pz = blkdiag(Px, Qw, Rv); % Covariance for extended state.
    [x_sigma, w_sigma, v_sigma] = get_sigma(Pz, xhat(:,1), pars);

    for t = 1:(Nsim + 1)
        % Take measurement.
        y_sigma = NaN(Ny, Nsigma);
        for i = 1:Nsigma
            y_sigma(:,i) = full(pars.h(x_sigma(:,i))) + v_sigma(:,i);
        end
        yhat(:,t) = y_sigma*pars.weights;
        e(:,t) = y(:,t) - yhat(:,t);

        % Apply correction.
        Ex = bsxfun(@minus, x_sigma, xhat(:,t));
        Ey = bsxfun(@minus, y_sigma, yhat(:,t));

        Exy = Ex*W*Ey';
        Eyy = Ey*W*Ey';

        L = Exy/Eyy;

        Px = Px - L*Exy';
        xhat(:,t) = xhat(:,t) + L*e(:,t);

        % Stop if at end of simulation.
        if t > Nsim
            break
        end

        % Calculate new sigma points, and clip if necessary.
        Pz = blkdiag(Px, Qw, Rv);
        [x_sigma, w_sigma, v_sigma] = get_sigma(Pz, xhat(:,t), pars);
        if clipping
            [x_sigma, pars.weights] = ukf_scale(x_sigma, pars);
        end

        % Simulate sigma points and calculate new xhat and Px.
        for i = 1:Nsigma
            x_sigma(:,i) = full(pars.f(x_sigma(:,i), w_sigma(:,i)));
        end
        xhat(:,t + 1) = x_sigma*pars.weights;
        Ex = bsxfun(@minus, x_sigma, xhat(:,t + 1));
        Px = Ex*W*Ex';
    end

    % Make a plot.
    pars.reactorplot(x, xhat, y, yhat, pars.yclean, pars.Delta);
    set(gca(), 'ylim', [-1, 1.5]);

    % Form data table.
    data = [pars.t', x', xhat', y', yhat'];