function [x_sigma, weights] = ukf_scale(x_sigma, pars)
    % [x_sigma, weights] = ukf_scale(x_sigma, pars)
    %
    % Rescales the sigma points and adjusts weights so that each sigma point has
    % no negative components.
    %
    % x_sigma should be a matrix of column vectors with the first vector giving
    % the central point (which must always be feasible).
    x_center = x_sigma(:,1); % Central point.
    x_outer = x_sigma(:,2:end);

    % Choose theta to make each point feasible.
    dx = abs(bsxfun(@minus, x_outer, x_center));
    err = abs(min(x_outer, 0)); % Gets negative part.
    err(x_center < 0,:) = 0; % Can't fix indices where x_center is negative.
    dx(err == 0) = 1; % Avoid division by zero.
    theta = 1 - max(err./dx);

    % Rescale outer points.
    x_outer = bsxfun(@times, theta, x_outer) ...
              + bsxfun(@times, 1 - theta, x_center);
    x_sigma = [x_center, x_outer];

    % Calculate weights.
    rho = (2*pars.lambda - 1)/(pars.Nsigma - sum(theta));
    a = -rho/(2*(pars.Nz + pars.lambda));
    b = (1 + rho)/(2*(pars.Nz + pars.lambda));
    weights = [0, a*theta]' + b;