% This file simulates the Batch Reactor Example (3.2.1) from Haseltine
% & Rawlings (2005)
% MHE is implemented to this example and it is shown that it
% performs better than EKF and UKF when a poor prior is used.
mpc = import_mpctools();

% States: x = [Ca Cb Cc]'
% Total pressure is measured (C = RT*[1 1 1])
Nsim = 120;
pars = batchreactor(Nsim);
Nx = pars.Nx;
Ny = pars.Ny;
Nw = pars.Nw;
Nv = pars.Nv;
Nt = 10;
xsim = pars.xsim;
ysim = pars.ysim;

% Pick stage costs.
Q = pars.Q;
R = pars.R;
sig_w = pars.sig_w;
sig_v = pars.sig_v;

lfunc = @(w, v) (w'*w)/sig_w.^2 + (v'*v)/sig_v.^2;
l = mpc.getCasadiFunc(lfunc, [Nw, Nv], {'w', 'v'}, {'l'});

% Now build MHE solver.
mhepar = struct('Pinv', mpctools.spdinv(pars.P), 'x0bar', pars.xhat0);

N = struct('x', Nx, 'w', Nw, 'y', Ny, 't', Nt);

lb = struct('x', zeros(Nx, Nt + 1)); % Concentrations are nonnegative.
ub = struct('x', 10*ones(Nx, Nt + 1)); % Concentrations shouldn't be this high.

buildsolvertime = tic();
solver = mpc.nmhe('f', pars.f, 'l', l, 'h', pars.h, 'y', ysim(:,1:(Nt + 1)), ...
                  'N', N, 'lb', lb, 'ub', ub, 'par', mhepar, ...
                  'priorupdate', 'smoothing', 'verbosity', 0);

% Loop.
xhat = NaN(Nx, Nsim + 1);
yhat = NaN(Ny, Nsim + 1);
tic();
for t = 1:(Nsim + 1)
    % Get new measurement or extend horizon.
    if t > Nt + 1
        solver.newmeasurement(ysim(:,t));
    else
        solver.truncatehorizon(t - 1);
    end

    % Solve MHE problem and save state estimate.
    solver.solve();
    if ~isequal(solver.status, 'Solve_Succeeded')
        warning('Solver failure at time %d!', t);
        break
    end

    solver.saveestimate(); % Stores estimate to history and updates prior.
    xhat(:,t) = solver.history(1).xhat(:,end);
    yhat(:,t) = solver.history(1).yhat(:,end);
end

% Make a plot.
pars.reactorplot(xsim, xhat, ysim, yhat, pars.yclean, pars.Delta);

