% Closed-loop evolution of nonlinear system for N = 2.
sys = nonlinsys();

% Define objective function and bounds.
Nt = 2;
Nu = sys.Nu;
Nx = sys.Nx;
V = @(x, u) sys.Vf(x);
for t = (Nt - 1):-1:0
    i = (t*Nu + 1):((t + 1)*Nu); % Index of current u.
    V = @(x, u) sys.l(x, u(i)) + V(sys.f(x, u(i)), u);
end

ulb = repmat(sys.ulb, Nt, 1);
uub = repmat(sys.uub, Nt, 1);
I = cell(Nu, 1);
for i = 1:length(I)
    I{i} = (i:Nu:(Nt*Nu))';
end


% Simulate closed loop for different numbers of iterations.
Nsim = 10;
Ps = [3, 10];
options = struct();
options.p3 = {'alphamax', 50, 'beta', 0.95};

x0 = [3; -3];
data = struct();
for P = Ps
    % Get options.
    key = sprintf('p%d', P);
    opts = structGet(options, key, {});

    % Preallocate simulation vectors.
    x = NaN(Nx, Nsim + 1);
    x(:,1) = x0;
    u = NaN(Nu, Nsim);

    % Generate an initial guess using the linear control law.
    [xguess, uguess] = warmstart(sys, x0, NaN(Nu, Nt), 'linear');

    % Simulate in closed loop.
    for t = 1:Nsim
        % Perform optimization.
        [useq, uiter] = distributedgradient(@(u) V(x(:,t), u), uguess(:), ...
                                            ulb, uub, 'I', I, 'Niter', P, ...
                                            opts{:});
        u(:,t) = useq(1:Nu);

        % Simulate system and update warm start.
        x(:,t + 1) = sys.f(x(:,t), u(:,t));
        uguess = [reshape(useq((Nu + 1):end), Nu, Nt - 1), NaN(Nu, 1)];
        [xguess, uguess] = warmstart(sys, x(:,t + 1), uguess, 'nonlinear');

        % Stop if something went wrong.
        if norm(x(:,t + 1)) > 100
            warning('x has diverged!');
            break
        end
    end

    % Make a plot.
    figure();
    t = 0:Nsim;

    subplot(2, 1, 1);
    hold('on');
    plot(t, x(1,:), '-or');
    plot(t, x(2,:), '-sb');
    title(sprintf('p = %d', P));

    subplot(2, 1, 2);
    hold('on');
    stairs(t, u(1,[1:end,end]), '-or');
    stairs(t, u(2,[1:end,end]), '-sb');
    plot([0, Nsim], ulb(1)*[1, 1], ':k', [0, Nsim], uub(1)*[1, 1], ':k');

    % Store data set.
    data.(key) = struct('t', t, 'x', x, 'u', u, 'ulb', ulb, 'uub', uub);
end

