function regions = solvempqp(prob, varargin)
% regions = solvempqp(prob, 'key', value, ...)
%
% Solves the mpQP given by
%
%   min  0.5*z'*Q*z + p'*C'*z
%    z
%   s.t. A*z <= b + S*z
%
% for all p satisfying
%
%   E*p <= e
%
% with prob a struct containing fields Q, C, A, b, and S. Note that Q must be
% strictly positive definite. E and e can also be provided (they default to
% empty matrices).
%
% Available options are as follows:
%
% - 'display' : Whether to display status of iterations (default true)
% - 'printevery' : Number of iterations for each print (default 100)
% - 'logfile' : Filename to write detailed log (default '', which means no log)
% - 'vertices' : Whether to find vertices for each region (default true for
%                Np = 2, false for Np > 2)
% - 'plot' : Whether to make a plot of all the regions (default false; ignored
%            if 'vertices' is not true); only possible if length(p) == 2
% - 'maxiter' : Maximum integer number of iterations (default inf)
% - 'maxtime' : Maximum time to run (in seconds; default inf).
% - 'ascell' : Whether to return cell array (default true)
%
% Returned value is cell array of structs containing the halfspace and
% vertex representations of the polytopes. If 'ascell' is false, instead returns
% a struct whose fields are hexidecimal ID numbers and values contain the other
% information.
%
% Essential mpQP algorithm is from "A multiparametric quadratic programming
% algorithm with polyhedral computations based on nonnegative least squares"
% (Bemporad, 2015), although the polyhedral computations are performed using
% qhull and linprog rather than nonnegative least squares.
global SUPPRESS_OUTPUT
if nargin() < 1
    error('Argument prob is required!');
end
args = struct(varargin{:});
defaults = {'display', true(); 'logfile', ''; 'vertices', NaN(); ...
            'plot', false(); 'maxiter', inf(); 'ascell', true(); ...
            'maxtime', inf(); 'printevery', 100};
options = struct();
for i = 1:size(defaults, 1)
    f = defaults{i, 1};
    options.(f) = structGet(args, f, defaults{i, 2});
end
if ~isempty(options.logfile)
    logfile = fopen(options.logfile, 'w');
    log_(logfile); % Enables logging.
end
log_('Log started %s\n', datestr(now()));

% Disable output if a global SUPPRESS_OUTPUT is set (overrides user preference).
if ~isempty(SUPPRESS_OUTPUT) && SUPPRESS_OUTPUT
    options.display = false();
end

% Convert to nomenclature from Bemporad (2015).
pmpqp = struct(); % Primal mpQP.
pmpqp.G = prob.A;
pmpqp.W = prob.b;
pmpqp.S = prob.S;
pmpqp.Q = prob.Q;
pmpqp.Q = 0.5*(pmpqp.Q + pmpqp.Q'); % Matlab is picky about symmetry.
pmpqp.F = prob.C;
pmpqp.c = zeros(size(pmpqp.F, 1), 1);
pmpqp.Qhalf = chol(pmpqp.Q);
pmpqp.E = structGet(prob, 'E', zeros(0, size(prob.S, 2)));
if isempty(pmpqp.E)
    pmpqp.e = zeros(0, 1);
else
    if ~isfield(prob, 'e')
        error('prob.e must be provided if prob.E is given!');
    end
    pmpqp.e = prob.e;
end
if isnan(options.vertices)
    options.vertices = (size(prob.S, 2) == 2);
end

% Convert to dual problem (with fewer parameters).
dmpqp = struct(); % Dual mpQP.
dmpqp.H = pmpqp.G*(pmpqp.Q\pmpqp.G');
dmpqp.D = pmpqp.G*(pmpqp.Q\pmpqp.F) + pmpqp.S;
dmpqp.d = pmpqp.G*(pmpqp.Q\pmpqp.c) + pmpqp.W;

% Choose qhull options.
qhullopts = {'Qt', 'QbB', 'Pp'};
if size(prob.S, 2) >= 5
    qhullopts = [qhullopts, {'Qx'}];
end

% To start, we know no constraints are active at the origin.
Ncon = size(dmpqp.D, 1);
opensets = struct('N1', struct('active', false(Ncon, 1))); % All cons inactive.
opensets.N1.id = getcrid(opensets.N1.active);
nopen = 1;
nfound = 0;

regions = struct();
probedregions = struct(opensets.N1.id, []);

niter = 0;
starttime = cputime();
options.maxtime = options.maxtime + starttime;
while nopen > 0 && niter < options.maxiter && cputime() < options.maxtime
    niter = niter + 1;
    if options.display && mod(niter, options.printevery) == 0
        fprintf('Iteration %d (%d found, %d left to search).\n', ...
                niter, nfound, nopen);
    end
    log_('Iteration %d\n', niter);

    activestr = sprintf('N%d', nopen);
    nopen = nopen - 1;
    currentset = opensets.(activestr);
    opensets.(activestr) = [];
    crid = currentset.id;
    log_('    Getting %s\n', crid);
    [cr, okay] = getcriticalregion(pmpqp, dmpqp, currentset);
    if okay
        log_('    Found representation\n');
        [nonredundant, cr.A, cr.b] = removeredundantcon(cr.A, cr.b, cr.x, ...
                                                        [], qhullopts);
        cr.cind = cr.cind(nonredundant);
        cr.ctype = cr.ctype(nonredundant);
        if options.vertices
            cr.V = halfspace2vertex(cr.A, cr.b, cr.x);
        end
        regions.(crid) = cr;
        nfound = nfound + 1;

        % Try to find adjacent critical regions.
        log_('    Searching for adjacent regions\n');
        for i = 1:length(nonredundant)
            n = cr.cind(i);
            newactive = currentset.active;

            tryqpsearch = true();
            switch cr.ctype(i)
            case 'w'
                newactive(n) = true();
            case 'y'
                newactive(n) = false();
                tryqpsearch = false(); % QP doesn't help in this case.
            otherwise
                continue
            end
            keepgoing = true();
            while keepgoing
                % Check whether the new active combination has full rank.
                % If so, add it. Otherwise, try once to fix it.
                newcrid = getcrid(newactive);
                if ~isfield(probedregions, newcrid)
                    probedregions.(newcrid) = [];
                    [newfullrank, newHiihalf] = checkcrfullrank(dmpqp, newactive);
                    if newfullrank
                        newcr = struct('active', newactive, 'id', newcrid, ...
                                       'Hiihalf', newHiihalf);
                        [opensets, nopen] = addcr(opensets, nopen, newcr);
                        keepgoing = false();
                    else
                        log_('        %s is not full-rank\n', newcrid);
                        if tryqpsearch
                            tryqpsearch = false();
                            log_('        Trying QP neighbor search\n');
                            cr.facet = i;
                            [newactive, keepgoing] = findadjacentcr(cr, pmpqp);
                            if ~keepgoing
                                log_('        No neighbor found\n');
                            end
                        else
                            keepgoing = false();
                        end
                    end
                else
                    keepgoing = false();
                end
            end
        end
    else
        log_('    %s is infeasible\n', crid);
    end
    log_('    %d regions open\n', nopen);
end
log_('Finished with %d open regions\n', nopen);
log_(); % Closes the log.
if options.display
    fprintf(['Found %d regions (%d left to search) after %d iterations (in ' ...
             '%g s).\n'], nfound, nopen, niter, cputime() - starttime);
end

% Make a plot.
if options.plot && size(pmpqp.S, 2) == 2
    fprintf('Making plot (may take some time).\n');
    figure();
    hold('on');
    fields = fieldnames(regions);
    colors = viridiscolors(length(fields));
    [~, cperm] = sort(sin(1:length(fields)));
    colors = colors(cperm,:); % (Pseudo-)randomize colors.
    for i = 1:length(fields)
        f = fields{i};
        fill(regions.(f).V(:,1), regions.(f).V(:,2), colors(i,:));
    end
    xlabel('p_1');
    ylabel('p_2', 'rotation', 0);
end

% Reshape into cell array.
if options.ascell
    regionsstruct = regions;
    fields = fieldnames(regionsstruct);
    regions = cell(length(fields), 1);
    for i = 1:length(fields)
        f = fields{i};
        regions{i} = regionsstruct.(f);
        regions{i}.id = f;
    end
end

end%function

function [fullrank, Hiihalf] = checkcrfullrank(dmpqp, activecon)
    % [fullrank, Hiihalf] = checkcrfullrank(dmpqp, activecon)
    %
    % Checks whether the critical region defined by the set of active
    % constraints is full-rank. If so, fullrank is true(), and the Cholesky
    % decomposition of the reduced Hessian is also returned. If not, fullrank is
    % false, and the second output is meaningless.
    narginchk(2, 2);
    i = activecon;
    Hii = dmpqp.H(i,i);
    if isempty(Hii)
        Hiihalf = Hii;
        fullrank = true();
    else
        [Hiihalf, flag] = chol(Hii);
        fullrank = (flag == 0) && all(diag(Hiihalf) > 1e-6);
    end
end%function

function [cr, okay] = getcriticalregion(pmpqp, dmpqp, crset)
    % [cr, okay] = getcriticalregion(pmpqp, dmpqp, crset)
    %
    % Returns corresponding critical region based on given set of active
    % constraints. cr is a struct with fields A, b, x (with x giving an interior
    % point of the critical region), cind (giving indices of w or y for each
    % row of the constraint matrix), and ctype (giving 'w' or 'y').
    %
    % If activecon does not defin an optimal set of constraints, then cr is
    % empty, and okay is false.
    narginchk(3, 3);

    % Check for cholesky decomposition in crset.
    activecon = crset.active;
    if isfield(crset, 'Hiihalf')
        Hiihalf = crset.Hiihalf;
        okay = true();
    else
        [okay, Hiihalf] = checkcrfullrank(dmpqp, activecon);
    end

    % Get inequalities that define critical region (often redundant).
    cr = struct();
    if okay
        i = activecon;
        j = ~activecon;
        Y = -(Hiihalf\(Hiihalf'\[dmpqp.D(i,:), dmpqp.d(i)]));
        W = dmpqp.H(j,i)*Y + [dmpqp.D(j,:), dmpqp.d(j)];
        P = [W; Y]; % Feasible region is P*[x; 1] >= 0.
        A = -P(:,1:end-1);
        b = P(:,end);

        % Screen out near-zero rows of A. In this case, we just set them to
        % actual zero so the LP solver can handle it.
        Anorms = max(abs(A), [], 2);
        badrows = (Anorms < 1e-10);
        if any(b(badrows) < 0)
            okay = false();
        end
        A(badrows,:) = 0;
        b(badrows) = 1;

        % Add the inequalities that define the region of interest for x.
        A = [A; pmpqp.E];
        b = [b; pmpqp.e];
        [A, b] = scalerows(A, b);
        Ne = length(pmpqp.e);

        % Attempt to find an interior point.
        if okay
            [x, okay] = findinteriorpoint(A, b);
            if okay
                % Critical region.
                cr.A = A;
                cr.b = b;
                cr.cind = [find(j); find(i); -1*(1:Ne)'];
                cr.ctype = [repmat('w', size(W, 1), 1); ...
                            repmat('y', size(Y, 1), 1); ...
                            repmat('e', Ne, 1)];
                cr.x = x;

                % Optimal solution.
                law = -pmpqp.Qhalf\(pmpqp.Qhalf'\(pmpqp.G(i,:)'*Y ...
                                                  + [pmpqp.F, pmpqp.c]));
                cr.Z = law(:,1:end-1);
                cr.z0 = law(:,end);
            end
        end
    end
end%function

function [z, feas, slack, active] = solveprimalqp(x, pmpqp)
    % Solves the primal qp at the given value of x using quadprog (Matlab)
    % or qp (Octave).
    narginchk(2, 2);
    Q = pmpqp.Q;
    f = pmpqp.F*x;
    A = pmpqp.G;
    b = pmpqp.S*x + pmpqp.W;
    Aeq = [];
    beq = [];
    if isOctave()
        zguess = zeros(size(Q, 1), 1);
        [z, ~, flag] = qp(zguess, Q, f, Aeq, beq, [], [], [], A, b);
        feas = (flag.info == 0);
    else
        prob = struct('H', Q, 'f', f, 'Aineq', A, 'bineq', b, ...
                      'Aeq', Aeq, 'beq', beq, 'options', ...
                      optimoptions('quadprog', 'display', 'off'), ...
                      'solver', 'quadprog');
        [z, ~, flag] = quadprog(prob);
        feas = (flag == 1);
        if isempty(z)
            z = NaN(length(f), 1);
        end
    end
    slack = pmpqp.G*z - pmpqp.S*x - pmpqp.W;
    active = (abs(slack) < 1e-5);
end%function

function [opensets, nopen] = addcr(opensets, nopen, newcr)
    % Adds the new set to the list of open sets.
    nopen = nopen + 1;
    openstr = sprintf('N%d', nopen);
    opensets.(openstr) = newcr;
    log_('        Adding %s\n', newcr.id);
end%function

function [newactive, feas] = findadjacentcr(cr, pmpqp)
    % Finds an adjacent CR by stepping in the direction of the shared facet and
    % solving the QP.
    narginchk(2, 2);

    % Get point on the facet.
    ineq = true(size(cr.A, 1), 1);
    ineq(cr.facet) = false();
    [x0, okay] = findinteriorpoint(cr.A(ineq,:), cr.b(ineq), ...
                                   cr.A(cr.facet,:), cr.b(cr.facet));
    if okay
        normal = cr.A(cr.facet,:)';
        x0p = x0 + 1e-4*normal/norm(normal);
        [~, feas, ~, newactive] = solveprimalqp(x0p, pmpqp);
    else
        feas = false();
    end
    if ~okay || ~feas
        newactive = [];
    end
end%function

function [A, b, scale] = scalerows(A, b, zerotol)
    % [A, b, scale] = scalerows(A, b, [zerotol])
    %
    % Scales the rows of A and b so that each row of [A, b] has unit norm.
    % If zerotol is given, any entries of A with absolute value less than
    % zerotol (after scaling) are set to zero. The default for zerotol is 1e-12.
    %
    % Any trivial rows (with A(i,:) = 0 and b(i) = 0) are set to b = 1.
    narginchk(2, 3);
    if nargin() < 3
        zerotol = 1e-12;
    end

    % Perform scaling.
    Z = [A, b];
    scale = sqrt(sum(Z.^2, 2));
    trivrows = (scale == 0);
    scale(trivrows) = 1; % Don't do anything to these guys.

    A = bsxfun(@rdivide, A, scale);
    b = b./scale;
    b(trivrows) = 1;

    % Screen out any elements that are very small.
    A(abs(A) < 1e-12) = 0;
end%function

function hex = bin2hex(bits)
    % Converts logical vector bits to hexidecimal string.
    hex = '0123456789ABCDEF';
    bits = bits(:);
    Npad = 4 - mod(length(bits), 4);
    if Npad == 4
        Npad = 0;
    end
    bits = [zeros(Npad, 1); double(bits)];
    Nhex = length(bits)/4;
    digits = [8, 4, 2, 1]*reshape(bits, 4, Nhex);
    hex = hex(digits + 1);
end%function

function crid = getcrid(active)
    % Returns the string ID of the given set of active constraints.
    crid = ['CR_', bin2hex(active)];
end%function

function log_(file_or_str, varargin)
    % Writes to log file using sprintf syntax. Pass integer file ID to start
    % logging. Call with no arguments to close.
    persistent logfile
    if nargin() == 1 && ~ischar(file_or_str) && isscalar(file_or_str)
        % Open log file.
        logfile = file_or_str;
        if logfile == -1
            error('Invalid log file!');
        end
    elseif nargin() == 0
        if ~isempty(logfile)
            fclose(logfile);
            logfile = [];
        end
    elseif ~isempty(logfile)
        fprintf(logfile, file_or_str, varargin{:});
        if isOctave()
            fflush(logfile);
        end
    end
end%function