function [mpqp, T] = ss2mpqp(varargin)
% mpqp = ss2mpqp(sys)
%
% Converts the linear state-space system described by sys.
%
% sys must contain the following fields:
%
% - A, B: matrices that describe model x^+ = A*x + B*u
% - Q, R: matrices for stage cost x'*Q*x + u'*Q*u
% - S: cross-term matrix for 2*x'*S'u (defaults to 0).
% - P: matrix for terminal penalty x'*P*x (defaults to Q).
% - N: integer choosing the horizon
%
% sys can optionally contain the following fields:
%
% - xlb, xub, ulb, uub: vectors of bounds for x or u
% - C, D, ylb, yub: output matrices and bounds
% - Af, bf: matrices that define Xf as polyhedron Af*xf <= bf
% - keepstates: True or False whether to eliminate the state matrices. If True,
%               variables are interleaved as [u(0); x(1); u(1); x(2); ...].
% - dualstates: True or False whether to dualize the state equality constraints.
%               If True, variables are interleaved as [u(0); lam(1); x(1); ...]
%               Note that this implicitly sets keepstates=True.
%
% Values can also be passed as (key, value) pairs instead of as a struct.
%
% Note that all of these bounds must be chosen so that the feasible set is
% full-dimensional. In particular, lb < ub, and (Af, bf) cannot contain
% implicit equality constraints.
%
% Returned value is a struct with fields to describe the mpQP
%
%     min  0.5*z'*Q*x + p'*C*z
%      z
%     s.t. A*z <= b + S*p
%
% with p = x(0) and z = [u(0); u(1); ...].
if nargin() == 1
    sys = varargin{1};
else
    sys = struct(varargin{:});
end
if ~isstruct(sys)
    error('Input sys must be a struct!');
elseif any(~isfield(sys, {'A', 'B', 'Q', 'R', 'N'}))
    error('Input sys is missing required fields!');
end

% Get main matrices and sizes.
A = sys.A;
B = sys.B;
Q = sys.Q;
R = sys.R;
[Nx, Nu] = size(B);
Nt = sys.N;

% Get optional values.
S = structGet(sys, 'S', zeros(Nx, Nu));
P = structGet(sys, 'P', Q);
xlb = structGet(sys, 'xlb', -inf(Nx, 1));
xub = structGet(sys, 'xub', inf(Nx, 1));
ulb = structGet(sys, 'ulb', -inf(Nu, 1));
uub = structGet(sys, 'uub', inf(Nu, 1));
C = structGet(sys, 'C', zeros(0, Nx));
Ny = size(C, 1);
D = structGet(sys, 'D', zeros(Ny, Nu));
ylb = structGet(sys, 'ylb', -inf(Ny, 1));
yub = structGet(sys, 'yub', inf(Ny, 1));
Af = structGet(sys, 'Af', zeros(0, Nx));
if isempty(Af)
    bf = zeros(0, 1);
else
    if ~isfield(sys, 'bf')
        error('Must provide bf if Af is given!');
    else
        bf = sys.bf;
    end
end
dualstates = structGet(sys, 'dualstates', false());
keepstates = dualstates || structGet(sys, 'keepstates', false());

% Compute state transition matrices that give
%
%   w = T*z
%
% with z = [x(0); u(0); u(1); ...; u(N - 1)], and
% w = [x(0); u(0); x(1); u(1); ...; x(N - 1); u(N - 1); x(N)].
Nxu = Nx + Nu;
Nw = Nt*Nxu + Nx;

% Rows corresponding to x(k).
ix = bsxfun(@plus, (1:Nx)', Nxu:Nxu:(Nxu*Nt));

% Columns for x(0) and u(k).
jx0 = (1:Nx)';
ju = bsxfun(@plus, (1:Nu)', Nx:Nxu:(Nxu*Nt));

% Get state transition matrix.
T = eye(Nw);
if keepstates
    Nz = Nw;
    keep = true(Nw, 1);
else
    Cont = zeros(Nx, Nw);
    Cont(:,jx0) = eye(Nx);
    for k = 1:Nt
        % Update controllability matrix.
        Cont = A*Cont;
        Cont(:,ju(:,k)) = B;

        % Add entries for x(k).
        T(ix(:,k),:) = Cont;
    end
    Nz = Nt*Nu + Nx;
    keep = [jx0(:); ju(:)];
end
T = T(:,keep);

% Get model constraints.
if keepstates
    Aeq = kron(eye(Nt + 1), [A, B]) + kron(diag(ones(Nt, 1), 1), ...
                                           [-eye(Nx), zeros(size(B))]);
    Aeq = Aeq(1:(end - Nx),1:(end - Nu));
    beq = zeros(Nt*Nx, 1);
else
    Aeq = zeros(0, Nw);
    beq = zeros(0, 1);
end

% Get output constraints.
A_y = [kron(eye(Nt), [C, D]), zeros(Ny*Nt, Nx)];
A_y = [A_y; -A_y];
b_y = [repmat(yub, Nt, 1); repmat(-ylb, Nt, 1)];

% Get polyhedron for bound constraints.
lb = [-inf(Nx, 1); repmat([ulb; xlb], Nt, 1)];
ub = [inf(Nx, 1); repmat([uub; xub], Nt, 1)];
[A_lim, b_lim] = hyperrectangle(lb, ub);

% Get polyhedron for terminal constraints.
A_Xf = [zeros(size(Af, 1), Nxu*Nt), Af];
b_Xf = bf;

% Formulate objective function.
H = blkdiag(kron(eye(Nt), [Q, S; S', R]), P);

% Reduce all of the matrices using the state transition matrix.
Alt = [A_lim; A_y; A_Xf]*T;
blt = [b_lim; b_y; b_Xf];
Aeq = Aeq*T;
H = T'*H*T;

% Dualize if requested.
if dualstates
    [H, Alt] = dualize(H, Alt, Aeq, [Nx, Nx + Nu]);
    Nz = Nz + Nx*Nt;
    Aeq = zeros(0, Nz);
end

% Remove trivial rows.
keep = (blt ~= inf());
Alt = Alt(keep,:);
blt = blt(keep);

% Formulate QP matrices.
mpqp = struct();

x0 = [true(Nx, 1); false(Nz - Nx, 1)];

mpqp.A = Alt(:,~x0);
mpqp.b = blt;
mpqp.S = -Alt(:,x0);

mpqp.Aeq = Aeq(:,~x0);
mpqp.beq = beq;
mpqp.Seq = -Aeq(:,x0);

mpqp.Q = H(~x0,~x0);
mpqp.C = H(~x0,x0);

end%function

function [H_, Alt_] = dualize(H, Alt, Aeq, blocksize)
    % Dualizes the equality constraints for the given QP.
    %
    % blocksize should give the size of each variable block in Aeq. Dual
    % multipliers are then interleaved between the blocks.
    Nvar = size(H, 1);
    Nblocks = size(Aeq, 1)/blocksize(1);
    if round(Nblocks) ~= Nblocks
        error('blocksize(1) must be a multiple of size(Aeq, 1)!');
    end

    % Get logical masks for variables and dual multipliers.
    var = [repmat([true(blocksize(2), 1); ...
                   false(blocksize(1), 1)], Nblocks, 1); ...
           true(Nvar - blocksize(2)*Nblocks, 1)];
    mul = ~var;

    Ntot = length(var);

    H_ = zeros(Ntot, Ntot);
    H_(var,var) = H;
    H_(mul,var) = Aeq;
    H_ = 0.5*(H_' + H_); % Make symmetric.

    Alt_ = zeros(size(Alt, 1), Ntot);
    Alt_(:,var) = Alt;
end%function
