% Example explicit MPC for a 2D system From "Optimal control of constrained
% piecewise affine discrete-time systems" (Mayne and Rakovic, 2002).




% Choose system.
A = [1,0.1; 0,1];
B = [0; 0.0787];
Q = [1, 0; 0, 0];
R = 0.1;
ulb = -1;
uub = 1;
xlb = -inf(2, 1);
xub = inf(2, 1);
N = 20;
[Nx, Nu] = size(B);

% Find LQR terminal set.
[K, P] = dlqr(A, B, Q, R);
K = -K;
[Xn, Xnverts] = findXn(A, B, K, N, xlb, xub, ulb, uub, 'lqr');
A_lqr = Xn{1}.A;
b_lqr = Xn{1}.b;
feas = Xn{end}; % Gives region X_N.

% Get mpQP, solve, and plot.
mpqp = ss2mpqp('A', A, 'B', B, 'Q', Q, 'R', R, 'P', P, 'N', N, ...
               'ulb', ulb, 'uub', uub, 'Af', A_lqr, 'bf', b_lqr);
regions = solvempqp(mpqp, 'plot', false());


xbox = [-6, 6; -3, 3];
Npts = 10000;
Nmax = 20000;
xpts = NaN(Nx, Npts);
lookuptimes = NaN(Npts, 1);
qptimes = NaN(Npts, 1);
jx = 0;
rand('state', 917);
for i = 1:Nmax
    xmin = xbox(:,1);
    dx = xbox(:,2) - xbox(:,1);
    xtry = xmin + dx.*rand(Nx, 1);
    if all(feas.A*xtry - feas.b <= 0)
        jx = jx + 1;
        if mod(jx, 500) == 1
            fprintf('Point %d of %d\n', jx, Npts);
        end
        xpts(:,jx) = xtry;

        % Do PWA lookup.
        tic();
        [zlookup, id] = pwalookup(xtry, regions, feas);
        lookuptimes(jx) = toc();

        % Solve QP.
        tic();
        H = mpqp.Q;
        q = mpqp.C*xtry;
        A = mpqp.A;
        b = mpqp.b + mpqp.S*xtry;
        if isOctave()
            z0 = zeros(Nu*N, 1);
            [zqp, ~, status] = qp(z0, H, q, [], [], [], [], [], A, b);
            okay = (status.info == 0);
        else
            H = 0.5*(H + H'); % Matlab is picky.
            options =  optimoptions('quadprog', 'display', 'off', ...
                                    'OptimalityTolerance', 1e-15, ...
                                    'StepTolerance', 1e-15, ...
                                    'MaxIterations', 500);
            prob = struct('H', H, 'f', q, 'Aineq', A, 'bineq', b, ...
                          'solver', 'quadprog', 'options', options);
            [zqp, ~, status] = quadprog(prob);
            okay = (status == 1);
        end
        if ~okay
            warning('QP failed at step %d!', jx);
        end
        qptimes(jx) = toc();

        % Make sure the solutions agree.
        if max(abs(zqp - zlookup)) > 1e-5
            warning('Mismatch for point %d!', jx);
        end

        % Stop if we've taken enough points.
        if jx == Npts
            break
        end
    end
end
fprintf('Sampled %d points.\n', jx);
xpts = xpts(:,1:jx);
lookuptimes = 1000*lookuptimes(1:jx); % Convert to ms.
qptimes = 1000*qptimes(1:jx); % Conver to ms.


sig = 1; % Kernel width in ms.
tplot = linspace(0, 100, 1001); % Plot range in ms.

lookupdensity = kerneldensity(lookuptimes, sig, tplot);
qpdensity = kerneldensity(qptimes, sig, tplot);

figure();
plot(tplot, lookupdensity, '-r', tplot, qpdensity, '-g');
legend('Explicit', 'Implicit', 'Location', 'NorthEast');
title(sprintf('%d Samples', size(xpts, 2)));
xlabel('Solution Time (ms)');
ylabel('Samples');

scatterplot(xpts, lookuptimes, tplot, 'Explicit');
scatterplot(xpts, qptimes, tplot, 'Implicit');

