% Accuracy of explicit numerical integration of a linear ODE.

% Butcher tableaus (only a and b) for the three methods.
tabs = struct();
tabs.euler = struct('a', 0, 'b', 1);
tabs.heun = struct('a', [0, 0; 1, 0], 'b', [0.5, 0.5]);
tabs.rk4 = struct('a', [0, 0, 0, 0; 0.5, 0, 0, 0; 0, 0.5, 0, 0; 0, 0, 1, 0], ...
                  'b', [1/6, 1/3, 1/3, 1/6]);
tabnames = fieldnames(tabs);

% Run each method.
A = [0, 1; -1, 0];
f = @(x) A*x;
x0 = [1; 0];

% Choose values of M to test. Note that the figure in the text goes up to
% M = 10^6, which takes about an hour to run.
Ms = [2, 32, 1e2, 1e3, 1e4];
err = struct();
for n = 1:length(tabnames)
    n = tabnames{n};
    err.(n) = NaN(size(Ms));
    fprintf('Running %s.\n', n);
    for i = 1:length(Ms)
        M = Ms(i);
        fprintf('  M = %d\n', M);
        a = tabs.(n).a;
        b = tabs.(n).b;
        N = ceil(M/length(b)); % Number of integration steps.
        h = 2*pi()/N;
        x = x0;
        for t = 1:N
            x = butcherintegration(x, f, h, a, b);
        end
        err.(n)(i) = norm(x - x0);
    end
end

% Make a plot.
figure();
hold('on');
colors = {'r', 'g', 'b'};
for n = 1:length(tabnames)
    c = colors{n};
    n = tabnames{n};
    loglog(Ms, err.(n), ['-o', c], 'displayname', n);
end
xlabel('collocation points M');
ylabel('Error');
legend('location', 'northwest');

