% Illustration of GL4 collocation approximation

% Butcher tableau for GL4.
tab = struct('a', [1/4, 1/4 - sqrt(3)/6; 1/4 + sqrt(3)/6, 1/4], ...
             'b', [1/2, 1/2], ...
             'c', [1/2 - sqrt(3)/6, 1/2 + sqrt(3)/6]);

% Run one step of GL4.
h = 2*pi()/5;
x0 = [1; 0];
[x1, ~, ~, k] = butcherintegration(x0, @stiffode, h, tab.a, tab.b);

% Get exact solution.
t = linspace(0, h, 1001);
x = [cos(t); -sin(t)];
xdot = [-sin(t); -cos(t)];

% Get collocation polynomial approximation.
k1 = k(:,1);
k2 = k(:,2);
c1 = tab.c(1);
c2 = tab.c(2);

rho1 = h*(k1*c2 - k2*c1)/(c1 - c2);
rho2 = 0.5*h*(k1 - k2)/(c1 - c2);

xc = NaN(2, length(t));
xdotc = NaN(2, length(t));
xnode = NaN(2, length(tab.c));
for i = 1:2
    xc(i,:) = x0(i) - rho1(i)*(t/h) + rho2(i)*(t/h).^2;
    xdotc(i,:) = -rho1(i)/h + 2*rho2(i)/h*(t/h);
    xnode(i,:) = x0(i) - rho1(i)*tab.c + rho2(i)*tab.c.^2;
end

% Get tangent lines for derivative.
dt = 0.03; % Width for tangents.
tt = [tab.c(1) + [-dt, dt], NaN(), tab.c(2) + [-dt, dt]];
xt = [xnode(:,1) - h*k1*dt, xnode(:,1) + h*k1*dt, NaN(2, 1), ...
      xnode(:,2) - h*k2*dt, xnode(:,2) + h*k2*dt];

% Make a plot.
figure();
t = t/h;
for i = 1:2
    subplot(2, 2, i);
    hold('on');
    plot(t, x(i,:), '-k', t, xc(i,:), '--k');
    plot(tt, xt(i,:), '-k', 'linewidth', 5);
    ylabel(sprintf('x_%d', i), 'rotation', 0);
    xlabel('\tau/h');

    subplot(2, 2, 2 + i);
    hold('on');
    plot(t, xdot(i,:), '-k', t, xdotc(i,:), '--k');
    plot(tab.c, [k1(i), k2(i)], 'ok', 'markerfacecolor', 'k');
    ylabel(sprintf('dx_%d/dt', i), 'rotation', 0);
    xlabel('\tau/h');
end

% Save data.
data = struct('t', t, 'x', x, 'xdot', xdot, 'xc', xc, 'xdotc', xdotc, ...
              'tt', tt, 'xt', xt, 'h', h, 'k', k, 'c', tab.c, 'xn', xnode);
save('-v7', 'gl4illustration.mat', '-struct', 'data');
